using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data.Linq;
using System.Data.Linq.Mapping;
using System.Reflection;
using System.Configuration;
using People.DB;

/*
Primary Objects Linquify - Business Class Generator
Copyright (C) 2010 Kory Becker
http://www.primaryobjects.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.
*/

namespace People.Types.GeneratedTypes.BaseTypes
{
    public abstract class BusinessBaseType<TEntity, TContext>
        where TEntity : class, new()
        where TContext : DataContext, new()
    {
        protected TContext Context = new TContext();
		private Exception _lastException;
		
        protected BusinessBaseType()
        {
			Context.Connection.ConnectionString = ConfigurationManager.ConnectionStrings["MyConnectionString"].ConnectionString;
            Context.DeferredLoadingEnabled = false;
        }

        protected TEntity Load<TValue>(TValue pk)
        {
			try
			{
				Type typeOfEntity = typeof(TEntity);

				var primaryKey = (from p in typeOfEntity.GetProperties()
								from attr in p.GetCustomAttributes(false)
								where attr is ColumnAttribute && (attr as ColumnAttribute).IsPrimaryKey
								select p).FirstOrDefault();

				MetaTable metaTable = Context.Mapping.GetTable(typeof(TEntity));
				string tableName = metaTable.TableName;
	            
		        string Sql = "SELECT " + GetPropertyList(typeOfEntity) + " FROM " + tableName + " WHERE " + primaryKey.Name + "={0}";

				IEnumerable<TEntity> entityList = Context.ExecuteQuery<TEntity>(Sql, pk);
	            
				TEntity entity = entityList.Single();
				return entity;
            }
            catch (Exception excep)
            {
				_lastException = excep;
				return null;
            }
        }

        protected TEntity Delete<TValue>(TValue pk)
        {
            TEntity entity = Load<TValue>(pk);

			if (entity != null)
			{
				Table<TEntity> table = Context.GetTable(typeof(TEntity)) as Table<TEntity>;
				table.DeleteOnSubmit(entity);

				Context.SubmitChanges();
			}
			else
			{
				throw new Exception("Error in Delete(): Verify the primary key property has been set on the object before calling.\n" + _lastException.Message);
			}
			
            return entity;
        }

        protected bool Save<TValue, TBusiness>(TBusiness businessObject, TValue pk, bool explicitInsert)
        {
            TEntity entity = new TEntity();

            // Get the name of the Primary Key.
            Type typeOfEntity = typeof(TEntity);
            var primaryKey = (from p in typeOfEntity.GetProperties()
                            from attr in p.GetCustomAttributes(false)
                            where attr is ColumnAttribute && (attr as ColumnAttribute).IsPrimaryKey
                            select p).FirstOrDefault();

            bool isNew = explicitInsert;
            object primaryKeyValue = Converter.GetPropertyValue(businessObject, primaryKey.Name);

            if (primaryKey.PropertyType.Name.ToLower() == "int32")
            {
                if ((int)primaryKeyValue == 0)
                {
                    isNew = true;
                }
            }
            else if (primaryKey.PropertyType.Name.ToLower() == "guid")
            {
                if ((Guid)primaryKeyValue == Guid.Empty)
                {
                    isNew = true;

                    // Auto-generate a GUID.
					PropertyInfo pi = businessObject.GetType().GetProperty(primaryKey.Name);
					pi.SetValue(businessObject, Guid.NewGuid(), null);
                }
            }
            else if (primaryKey.PropertyType.Name.ToLower() == "string")
            {
                if ((string)primaryKeyValue == "")
                {
                    isNew = true;
                }
            }

            if (isNew)
            {
                Table<TEntity> table = Context.GetTable(typeof(TEntity)) as Table<TEntity>;
                table.InsertOnSubmit(entity);
            }
            else
            {
                // Updating an existing entity. First, load the LINQ context to update.
                entity = Load<TValue>(pk);
                if (entity == null)
                {
					throw new Exception("Error in Save(): Verify the WhereClause property has been set before updating.\n" + _lastException.Message);
                }
            }

            // Set the new values.
            Converter.ConvertToBusinessObject<TEntity>(entity, businessObject);

            Context.SubmitChanges();
            
            // If inserting, load primary key value into businessObject to return.
            if (isNew)
            {
                object primaryKeyValueLinq = Converter.GetPropertyValue(entity, primaryKey.Name);

                PropertyInfo pi = businessObject.GetType().GetProperty(primaryKey.Name);
                pi.SetValue(businessObject, primaryKeyValueLinq, null);
            }
            
            return true;
        }

        protected List<TEntity> Find<TBusinessWhere>(TBusinessWhere entity)
        {
            // Form a dynamic LINQ query using the LINQ object.
            string MyWhere = "", MyTop = "", MyOrderBy = "";

            FieldInfo fi = entity.GetType().GetField("MyTop");
            if (fi != null)
            {
                MyTop = (string)fi.GetValue(entity);
            }

            fi = entity.GetType().GetField("MyWhere");
            if (fi != null)
            {
                MyWhere = (string)fi.GetValue(entity);
            }

            fi = entity.GetType().GetField("MyOrderBy");
            if (fi != null)
            {
                MyOrderBy = (string)fi.GetValue(entity);
            }

            Type typeOfEntity = typeof(TEntity);

            string sql = "SELECT " + MyTop + " " + GetPropertyList(typeOfEntity) + " FROM ";

            List<object> propertyValues = new List<object>();
            List<bool> propertyValuesSet = new List<bool>();
            int propertyUseCount = 0;

            MetaTable metaTable = Context.Mapping.GetTable(typeof(TEntity));
            string tableName = metaTable.TableName;

            sql += tableName + " WHERE ";

            if (MyWhere.Length > 0)
            {
                sql += MyWhere.Replace("'", "''") + " AND ";
                propertyUseCount = 1;
            }

            sql += "1=1";

            // First, get the boolean indicator values so we know which fields to use.
            FieldInfo[] fieldsInfo = entity.GetType().GetFields();
            foreach (FieldInfo field in fieldsInfo)
            {
                string propertyName = field.Name;
                object propertyValue = field.GetValue(entity);

                if (propertyName.EndsWith("Set") && propertyValue.GetType().Name.ToLower() == "boolean")
                {
                    // Boolean flag indicating if we should use this field or not. Allows us to use null in query.
                    propertyValuesSet.Add((bool)propertyValue);
                }
            }

            // Next, go through the actual properties and if the boolean indicator is true, use it in the query.
            int propertyIndex = 0;
            PropertyInfo[] propInfo = entity.GetType().GetProperties();
            foreach (PropertyInfo property in propInfo)
            {
                string propertyName = property.Name;
                object propertyValue = null;

                propertyValue = property.GetValue(entity, null);

                // Actual property value to use in query.
                if (propertyValuesSet[propertyIndex] == true)
                {
                    // This value was set by the user. Use this value in the query.
                    sql += " AND " + propertyName;

                    if (propertyValue == null)
                    {
                        sql += " IS NULL";
                    }
                    else
                    {
                        sql += "={" + propertyValues.Count + "}";
                        propertyValues.Add(propertyValue);
                    }

                    propertyUseCount++;
                }

                propertyIndex++;
            }

            if (MyOrderBy.Length > 0)
            {
                sql += " " + MyOrderBy;
            }

            List<TEntity> entityList = new List<TEntity>();

            if (propertyUseCount > 0)
            {
                entityList = Context.ExecuteQuery<TEntity>(sql, propertyValues.ToArray()).ToList();
            }

            return entityList;
        }

        protected List<TEntity> ToList()
        {
            Type typeOfEntity = typeof(TEntity);

            Table<TEntity> table = Context.GetTable(typeof(TEntity)) as Table<TEntity>;
            return table.ToList();
        }

        private string GetPropertyList(Type typeOfEntity)
        {
            // Create an explicit list of properties for a type to select, rather than using a *, which may be slower.
            string Sql = "";

            var properties = from p in typeOfEntity.GetProperties()
                             from attr in p.GetCustomAttributes(false)
                             where attr is ColumnAttribute
                             select p;

            int propIndex = 0;
            foreach (var prop in properties)
            {
                if (propIndex++ > 0)
                {
                    Sql += ", ";
                }

                Sql += prop.Name;
            }

            return Sql;
        }
    }

    public class Converter
    {
        public static object ConvertToBusinessObject<TEntity1>(TEntity1 originalObject, object abusinessType)
        {
            // Convert originalObject to a businessType object.
            object businessObject = null;
            Type businessType = originalObject.GetType();

            // Find the class
            Type obj = Type.GetType(businessType.AssemblyQualifiedName);

            // Get it's constructor
            ConstructorInfo constructor = obj.GetConstructor(new Type[] { });

            // Invoke it's constructor, which returns an instance of businessType.
            businessObject = constructor.Invoke(null);

            PropertyInfo[] propInfo = abusinessType.GetType().GetProperties();
            foreach (PropertyInfo property in propInfo)
            {
                if (!property.PropertyType.IsArray)
                {
                    SetPropertyValue(originalObject, property.Name, property.GetValue(abusinessType, null));
                }
            }

            return businessObject;
        }

        public static bool SetPropertyValue(object obj, string linqPropertyName, object propertyValue)
        {
            PropertyInfo[] propInfo = obj.GetType().GetProperties();
            foreach (PropertyInfo property in propInfo)
            {
                string propertyName = property.Name;

                if (!property.PropertyType.IsArray && propertyName.ToLower() == linqPropertyName.ToLower())
                {
                        object currentValue = property.GetValue(obj, null);
                        if (currentValue == propertyValue || (currentValue != null && currentValue.Equals(propertyValue)))
                        {
                            // Value has not changed.
                        }
                        else
                        {
                            property.SetValue(obj, propertyValue, null);
                        }

                    return true;
                }
            }

            return false;
        }

        public static object GetPropertyValue(object obj, string linqPropertyName)
        {
            PropertyInfo[] propInfo = obj.GetType().GetProperties();
            foreach (PropertyInfo property in propInfo)
            {
                string propertyName = property.Name;

                if (!property.PropertyType.IsArray && propertyName.ToLower() == linqPropertyName.ToLower())
                {
                    return property.GetValue(obj, null);
                }
            }

            return null;
        }
    }
}
